package Agent;

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import java.net.*;
import iaik.security.provider.IAIK;
import javax.crypto.*;
import java.security.*;


public abstract class Agent extends Thread 
                     implements java.io.Serializable {
  // Nome dell'agente
  protected String AgentName;  
  // Lista delle chiavi
  protected Hashtable KeyList;

  public Agent (String s) {
    super ("Agent - " + s);
    AgentName = getName();
  }

  public Agent() {
    super ("Agent");
    AgentName = getName();
  }

  final public String AgentName() {
    return AgentName;
  }


  // Metodo di recupero delle informazioni sulla classe dell'agente
  final public byte[] getClassBytes() {
    return getClassBytes ( getClass().getName() );
  }

  public byte[] getClassBytes (String className) {
    AgentClassLoader classLoader =
         (AgentClassLoader)(getClass().getClassLoader());
    // Prima di cercare la classe nel file system locale,
    // la si cerca nella tabella del class loader
    if ( classLoader != null ) {
      byte[] ClassBytes = classLoader.getClassBytes (className);
      if ( ClassBytes != null )
        return ClassBytes;
    }
    try {
      System.out.println ("Agent--> tento di caricare " + className);
      byte result[] = ClassBytesLoader.loadClassBytes (className);
      return result;
    } catch ( FileNotFoundException nof ) {
        nof.printStackTrace();
        return null;
    } catch ( Exception e ) {
        e.printStackTrace();
    }
    return null;
  }  

  // Metodo di migrazione dell'agente
  protected void migrate (String host, int port) 
                                     throws AgentException {
     Socket socket;
     try {
       System.out.println ("Agent--> Trasferimento su " + host + " : " + port);
       socket = new Socket (host, port);
       sendAgent (socket, host);
     } catch ( IOException e ) {
         throw new AgentException("Agent--> Impossibile aprire la socket");
     }
  }

  // Metodo per spedire in rete l'AgentPacket tramite la serializzazione	
  protected void sendAgent (Socket socket,String host) 
                        throws IOException, AgentException {    
    try {
      // Aggiungo il nuovo security provider
      Provider iaik = new iaik.security.provider.IAIK();
      Security.addProvider(iaik);     
      // Istanzio un nuovo cipher
      Cipher c = Cipher.getInstance("DES/ECB8/PKCS5Padding");
      // Inizializzo il cipher per la crittazione
      c.init (Cipher.ENCRYPT_MODE, (Key) KeyList.get(host) );
      ObjectOutputStream objOStream = new ObjectOutputStream (socket.getOutputStream());
      ObjectInputStream objIStream = new ObjectInputStream (socket.getInputStream());
      Boolean ok;
      System.out.println ("Agent--> Preparazione dati per migrazione...");
      // Calcolo dei tempi di migrazione : creazione pacchetto
      Date PackDate = new Date();
      AgentPacket agpack = createAgentPacket();
      System.out.println ("Agent--> Migrazione...");
      // Creazione di un SealedObject crittando e firmando l'AgentPacket
      SealedObject CryptPack = new SealedObject(agpack,c);
      // Calcolo dei tempi di migrazione : invio pacchetto
      Date InvioDate = new Date();
      System.out.println("Agent--> Tempo impiegato per la creazione del pacchetto: " + (InvioDate.getTime()-PackDate.getTime()) + " (msec)");
      // Invio sullo stream del pacchetto criptato
      objOStream.writeObject (CryptPack);
      ok = (Boolean)(objIStream.readObject());
      // Calcolo dei tempi totali di migrazione
      Date TotDate = new Date();
      System.out.println("Agent--> Tempo impiegato per la spedizione del pacchetto: " + (TotDate.getTime()-InvioDate.getTime()) + " (msec)");
      System.out.println("Agent--> Tempo totale impiegato: " + (TotDate.getTime()-PackDate.getTime()) + " (msec)");
      if ( ! ok.booleanValue() ) {
        throw new AgentException ("Agent--> Migrazione fallita");
      }
      System.out.println ("Agent--> Migrazione eseguita.");
    } catch ( ClassNotFoundException e ) {
        e.printStackTrace();
    } catch ( NoSuchPaddingException e ) {
        System.out.println("Agent-->Padding sbagliato"); 
    } catch ( NoSuchAlgorithmException ec ) {
        System.out.println("Agent--> Algoritmo sconosciuto"); 
    } catch ( Exception e ) {
        System.out.println("Agent--> Errore " + e.toString()); 
    }
  }

  // Metodo che crea il "pacchetto" agente
  protected AgentPacket createAgentPacket() {
    return new AgentPacket (this, getClassBytes());
  }

  // Metodo che esegue l'agente		  
  final public void run() {
    try {
      execute();
    } catch ( AgentException e ) {
	System.err.println ("Agent--> Eccezione non gestita " + e);
	e.printStackTrace();
    }
  }



  // Metodi da definire nelle classi derivate:

  // Metodo di inizio attivit dell'agente
  abstract protected void execute() throws AgentException;
  
  // Metodo per aggiornare le liste
  abstract public void SetListeNodo (ListaTabelle l, ListaTabelle le, Hashtable kl);
  
  // Metodo chiamato all'arrivo su un sito remoto
  abstract public void onArrival(AgentServer s) throws AgentException;
  
}
