package Agent;

import java.io.*;
import java.util.Hashtable;
import java.util.StringTokenizer;


// Class loader che non carica direttamente le classi dalla rete;
// quando necessarie, vengono cercate in una tabella hash interna al class loader
public class AgentClassLoader extends ClassLoader {

  private Hashtable classData = new Hashtable();

  public AgentClassLoader() {
  }

  // Metodo che aggiunge nella tabella hash della cache la classe
  synchronized public void addClassBytes (String className, byte classBytes[]) {
    if (classData.get (className) == null && classBytes != null)
      classData.put (className, classBytes);
  }

  public byte[] getClassBytes (String className) {
      return (byte[]) classData.get (className);
  }

  // Metodo per caricare le classi
  public Class loadClass(String className) throws ClassNotFoundException {
      return (loadClass(className, true));
  }
  
  protected synchronized Class loadClass(String className, boolean resolveIt)
                                   throws ClassNotFoundException {
      return loadClass (className, resolveIt, false);
  }

  protected synchronized Class loadClass (String className, boolean resolveIt,
                                            boolean force ) throws ClassNotFoundException {
     Class result;
     byte  classData[];
     // System.out.println ("ACL--> caricamento : " + className);
     
     /* proviamo prima con la cache */
     result = findLoadedClass(className);
     if (result != null) {
       // System.out.println ("ACL-->    --- recuperata dalla cache." );
       return result;
     }
     if ( ! force ) {
       try {
         result = super.findSystemClass(className);
         // System.out.println ("ACL-->    --- caricata dal file system locale");
         return result;
        } catch (ClassNotFoundException e) {
            // System.out.println ("ACL-->    --- non e' una classe di sistema");
        }
      }
     if ( className.startsWith ("java.") ) {
       // pericolo !!!
       throw new SecurityException (className);
     }
     
     /* proviamo a cercare nella nostra tabella */
     classData = getClassBytes(className);
     if (classData != null) {
       // System.out.println ("ACL-->    --- recupero dei bytes di " + className );
     } else {
         ClassNotFoundException e = new ClassNotFoundException (className);
         e.printStackTrace();
         throw e;
     }
     
     /* parsing */
     result = defineClass(className, classData, 0, classData.length);
     if (result == null) {
       throw new ClassFormatError();
     }
     if (resolveIt) {
       resolveClass(result);
     }
     System.out.println ("ACL-->    --- nuova classe caricata : " + className);
     return result;
 }

 public synchronized Class forceLoadClass (String className)
                                          throws ClassNotFoundException {
      return loadClass (className, true, true);
  }

}
