package Agent;

import java.lang.reflect.*;
import java.io.*;
import java.net.*;
import java.util.*;


public abstract class AgentEx extends Agent {
 
  protected Hashtable usedClasses = null;  
  // Indirizzo firmato per la certificazione
  protected byte FirmaBytes[];

  public AgentEx (String s) {
    super(s);
  }

  public AgentEx() {
    super();
  }

  // Metodo che setta la firma
  public void SetFirma (byte FirmaNodo[]) {
    FirmaBytes = FirmaNodo;
  }

  // Metodo di recupero delle informazioni sulla classe dell'agente
  synchronized public byte[] getClassBytes (String className) {
    byte[] ClassBytes = null;

    if ( usedClasses != null )
      ClassBytes = (byte[]) usedClasses.get(className);
    if ( ClassBytes != null ) {
      return ClassBytes;
    } else {
        ClassBytes = super.getClassBytes(className);
      }
    return ClassBytes;
  }

  // Metodo che crea un AgentPacketEx da spedire
  synchronized protected AgentPacket createAgentPacket() {
    return new AgentPacketEx (this, getClassBytes(), getUsedClasses(), FirmaBytes);
  }

  // Metodo che recupera le varie informazioni delle classi 
  // utilizzate dall'agente (sfruttando le API Reflection)
  synchronized protected Hashtable getUsedClasses() {
    if ( usedClasses == null )	
      // Se l'hash usedclasses e' gia' stato creato 
      // salta la sua creazione, e torna l'hashtable stesso
      createUsedClassesTable();

    return usedClasses;
  }

  // Metodo che provvede a riempire la tabella Hash
  synchronized protected void createUsedClassesTable() {
    if ( usedClasses != null )	{
      // Se l'hash usedclasses e' gia' stato creato ritorna subito 
      return;
    }
    usedClasses = new Hashtable();   // Crea l'hashtable
    Vector UsedClassesNames = new Vector();
    // Si ottengono "tutte" le classi utilizzate da questa classe
    getUsedClassesNames (getClass() , UsedClassesNames);
    byte[] classBytes;
    String className;
    Enumeration en = UsedClassesNames.elements();
    while ( en.hasMoreElements() ) {
      className = (String)en.nextElement();
      System.out.println ("classe: " + className);
      // Viene riempita la tabella Hash ottenendo i byte delle varie classi
      classBytes = getClassBytes (className);
      if ( classBytes != null ) {
      	usedClasses.put (className, classBytes);
	System.out.println ("Registrata " + className);
      }
    }
  }

  // Per il recupero dei nomi delle classi si adotta un algoritmo ricorsivo
  // Tramite le reflection API si ottengono le classi dei vari membri e dei
  // parametri e dei valori di ritorno dei vari metodi della classe specificata
  public Enumeration getUsedClassesNames() {
    if ( usedClasses != null ) 
      return usedClasses.keys();
    
    return null;
  }

  protected void getUsedClassesNames (Class c, Vector result) {
    Field[] fields = c.getDeclaredFields();
    Constructor[] constructors = c.getDeclaredConstructors();
    Method[] methods = c.getDeclaredMethods();
    Class[] declClasses = c.getDeclaredClasses();
    Class[] classes;
    int i, j;

    for (i = 0; i < fields.length; i++) {
      getUsedClassesNamesOf (fields[i].getType(), result);
    }
    for (i = 0; i < constructors.length; i++) {
      classes = constructors[i].getParameterTypes();
      if ( classes.length > 0 ) {
      	for (j = 0; j < classes.length; j++)
	   getUsedClassesNamesOf (classes[j], result);
      }
    }
    for (i = 0; i < methods.length; i++) {
      getUsedClassesNamesOf (methods[i].getReturnType(), result);
      classes = methods[i].getParameterTypes();
      if ( classes.length > 0 ) {
        for (j = 0; j < classes.length; j++)
          getUsedClassesNamesOf (classes[j], result);
      }
    }
    // getDeclaredClasses non  ancora implementata nel jdk
    // per adesso e' inutile (il vettore e' vuoto)   :-(
    for ( i = 0; i < declClasses.length; i++ ) {
      getUsedClassesNamesOf (declClasses[i], result);
    }
    getUsedClassesNamesOf (c.getSuperclass(), result);
  }

  protected void getUsedClassesNamesOf (Class classVar, Vector result) {
    String className = filter (classVar.getName());
    if ( isUsefulClass(className) && ! result.contains(className) ) {
      // E' la prima volta
      result.addElement(className);
      // Chiamata ricorsiva
      getUsedClassesNames(classVar, result);
    }
  }

  // Metodo che ripulisce da caratteri indesiderati (se e' un array ad esempio)
  protected String filter (String className) {
    if ( className.startsWith ("[L") )
      return className.substring (2);
   
    return className;
  }

  // Non vengono caricate le classi primitive e i relativi array
  // ed anche le classi dei package java e Agent
  final protected boolean isUsefulClass  (String className) {
    boolean result;
    result = (
      className.equals( "void" ) || 
      className.equals( "int" )  || className.equals( "[I" ) ||
      className.equals( "char" ) || className.equals( "[C" ) ||
      className.equals( "double" ) || className.equals( "[D" ) ||
      className.equals( "float" )  || className.equals( "[F" ) ||
      className.equals( "long" ) || className.equals( "[J" ) ||
      className.equals( "short" ) || className.equals( "[L" ) ||
      className.equals( "boolean" ) || className.equals( "[Z" ) ||
      className.equals( "byte" ) || className.equals( "[B" ) ||
      ( className.indexOf( '.' ) != -1 &&
        (className.startsWith("java") || className.startsWith("Agent")) )
    );
    return (! result);
  }
}

