// Agente di coordinamento

package Agent;

import java.io.*;
import java.util.*;

public class CoordAgent extends AgentEx {
  
  protected String host;
  protected int port;
  // Lista dei "riferimenti conosciuti"
  protected ListaTabelle ListaNodo;
  // Lista dei riferimenti errati alle tabelle 
  protected ListaTabelle ListaRifErrati;
  // Informazioni sul Nodo che ospita l'agente
  protected String HostLoc;
  protected int HostPort;
  // Il Server si pubblicizza tramite l'agente di coordinamento...
  protected static String pubHost;
  protected static int pubPort;   
  // Variabili per il Time to Live dell'agente
  protected int Time2Live;
  protected int NNodiVisitati;

  public CoordAgent(int t2l) {
    super ( "CoordAgent" );
    // Istanziazione delle liste 
    ListaNodo = new ListaTabelle();
    ListaRifErrati = new ListaTabelle();
    KeyList = new Hashtable() ;
    HostLoc="";
    HostPort=0;
    host = "";
    port = -1 ;
    Time2Live = t2l;
  }

  public CoordAgent (String s, int p, int t2l) {
    super ( "CoordAgent" );
    host = s;
    port = p;
    HostLoc="";
    HostPort=0;
    // Istanziazione delle liste 
    ListaNodo = new ListaTabelle();
    ListaRifErrati = new ListaTabelle();
    KeyList = new Hashtable() ;
    Time2Live = t2l;
  }


  // Metodo per aggiornare le liste	
  public void SetListeNodo (ListaTabelle SerLista, ListaTabelle SerLe, Hashtable SerKl) {

    // Se la dimensione di ListaNodo e' = 0, significa che  
    // il CoordAgent e' stato appena lanciato, 
    // quindi bisogna fare l'assegnamento della lista dei riferimenti errati
    if ( ListaNodo.size() == 0 ) 
      ListaRifErrati = SerLe;
    System.out.println ("Agent--> Liste ricevute - controllo e aggiornamento...");
    // Controllo ed elimino riferimenti errati al nodo del server..
    for (int i=0; i<ListaNodo.size(); i++) {
    EntryLista en = (EntryLista) ListaNodo.elementAt(i);
       if ( (en.IDTabella.compareTo("*") != 0) &&
            (en.Locazione.compareTo(HostLoc) == 0 ) &&
            (en.Porta == HostPort ) ) {
         // Nel caso che il riferimento sia AD UNA TABELLA su questo server 
         // ne controllo la validita' ed eventualmente lo elimino...
         if ( SerLista.SetTabella(en,0) == 0 ) {
           // Aggiunta del riferimento errato nella relativa lista...
           if ( ListaRifErrati.SetTabella((EntryLista) ListaNodo.elementAt(i)) == 0 )
             ListaRifErrati.addElement ((EntryLista) ListaNodo.elementAt(i));
           System.out.println ("Agent--> Eliminata la entry errata: ");
           System.out.println (en.IDTabella + " " + en.Locazione + " " + 
                               en.Porta + " " + en.Versione + " " + en.Data); 
           EntryLista temp = new EntryLista ("*", en.Locazione, en.Porta, en.Versione, en.Data);
           // ...se e' presente un altro riferimento generico a quel nodo
           // lo si cambia, altrimenti si elimina la entry dalla lista
           if ( ListaNodo.SetTabella(temp,0) == 0 ) {
             ListaNodo.setElementAt(temp,i);
           } else {
               ListaNodo.removeElementAt(i);
               // Aggiornamento eventuale del puntatore di migrazione...
               if ( i <= NNodiVisitati )
                 NNodiVisitati--;
             }
           i--;  // Avendo eliminato una entry decremento il contatore
         }
       }
    }
    // Controllo la presenza di riferimenti errati nella lista del server
    // Vengono totalmente eliminati se si sta "pubblicizzando" un'uscita
    // di un nodo dal sistema (host=="" , port<0)
    if ( port<0 ) {
      // Recupero la prima entry che si riferisce al nodo che e' uscito
      EntryLista NodoUscito = (EntryLista) ListaRifErrati.elementAt(0);
      SerLista.EliminaRif (NodoUscito.Locazione, NodoUscito.Porta);
    } 

    // Controllo incrociato delle liste
    SerLista.ConfrontaListe (ListaNodo, ListaRifErrati, SerLe);

    // Aggiornamento incrociato delle liste dei riferimenti errati
    SerLe.ConfrontaListe (ListaRifErrati);
    // Elimino riferimenti errati inutili al server stesso
    SerLe.EliminaRif (HostLoc, HostPort);

    // Aggiornamento delle chiavi
    for (int i=0; i<SerLista.size(); i++) {
      try {
        EntryLista ent = (EntryLista) SerLista.elementAt(i);
        SerKl.put (ent.Locazione, KeyList.get(ent.Locazione));
      } catch (Exception e) { }
    }
    KeyList = SerKl;
  }


  // Metodo che setta la firma
  public void SetFirma (byte FirmaNodo[]) {
    FirmaBytes = FirmaNodo;
  }


  // Metodo di inizio attivit dell'agente
  protected void execute() throws AgentException {
    System.out.println ("CA--> Salve sono l'agente " + AgentName());
    // Considera i nodi da visitare scartando il nodo di partenza (prima entry)
    // nel caso che debba fare una migrazione di ritorno (Host!="" , port<0)
    if (port<0) {
      NNodiVisitati = 0;
    } else {
        NNodiVisitati = 1;
      }
    System.out.println ("CA--> Questa e' la mia lista:");
    for (int i=0; i<ListaNodo.size(); i++) {
      EntryLista en;
      en = (EntryLista) ListaNodo.elementAt(i);
      System.out.println (en.IDTabella + " " + en.Locazione + " " +
                       en.Porta + " " + en.Versione + " " + en.Data);
    }
    if ( ListaRifErrati.size() != 0 ) {
      System.out.println ("CA--> Questa e' la mia lista dei riferimenti errati:");
      for (int i=0; i<ListaRifErrati.size(); i++) {
        EntryLista en;
        en = (EntryLista) ListaRifErrati.elementAt(i);
        System.out.println (en.IDTabella + " " + en.Locazione + " " +
                            en.Porta + " " + en.Versione + " " + en.Data);
      }
    } else {
        System.out.println ("CA--> La mia lista dei riferimenti errati e' vuota");
      }
    MigraSuNodo();	
  }


  // Metodo di migrazione dell'agente
  private void MigraSuNodo () {
    // Se non ci sono piu' nodi da visitare 
    // o il Time to Live  scaduto termina  
    while ((ListaNodo.size() > NNodiVisitati) && (Time2Live > 1)) { 
       // Tentativo di migrazione
       System.out.println ("CA--> Tento la migrazione..."); 
       EntryLista en;
       en = (EntryLista) ListaNodo.elementAt(NNodiVisitati);
       try {
         if (ListaNodo.CheckLocazione (en.Locazione, en.Porta, NNodiVisitati)) {
           // Non migra su un nodo gia' visitato, ma
           // incrementa comunque il numero di nodi visitati..
           NNodiVisitati++;
           System.out.println ("CA--> Su questo nodo ci sono gia' stato... : " + en.Locazione +
                               " : " + en.Porta);
         } else {
             try {
                HostLoc = en.Locazione;
                HostPort = en.Porta;
                migrate (en.Locazione, en.Porta);
                return;         
             } catch (AgentException ae) {
                // Elimina la entry del nodo che ha causato l'errore
                ListaNodo.removeElementAt(NNodiVisitati);
                System.out.print ("CA--> Eliminata la entry : "); 
                System.out.println (en.IDTabella + " " + en.Locazione + 
                   " " + en.Porta + " " + en.Versione + " " + en.Data);
             }
           }
        } catch (Exception e) { 
            System.out.println ("CA--> Errore durante il check della lista");   
        }
    }
    // L'ultima migrazione e' sul nodo di partenza
    // cioe' se Time2Live = 1 oppure siamo in fondo alla lista
    if ( ((Time2Live == 1) || (ListaNodo.size() == NNodiVisitati)) &&
         (port > 0) ) {
      try {
        System.out.println ("CA--> Tento la migrazione di ritorno..."); 
        HostLoc = host;
        HostPort = port;
        Time2Live = 1;
        migrate (host, port);
        return;         
      } catch (AgentException ae) {
          // Migrazione di ritorno fallita
          System.out.println ("CA--> Migrazione di ritorno fallita :-(( ");
      }
    }
  }   


  // Metodo chiamato all'arrivo su un sito remoto
  public void onArrival(AgentServer server) throws AgentException {

     // Aggiornamento delle liste a carico dell'agente...
     System.out.println ("CA--> Liste date all'agente per l'aggiornamento ");
     SetListeNodo (server.Lista, server.ListaEliminati, server.KeyList);
     // Si da' la firma all'agente...
     SetFirma (server.FirmaNodo);

     // Incremento del numero di nodi visitati
     NNodiVisitati++;
     // Decremento del Time to live
     Time2Live--;
     System.out.println ("Salve, sono l'agente " + AgentName() +
                         " appena arrivato..." ) ;
     // Visualizzazione della lista sul nodo di arrivo
     System.out.println ("CA--> Questa e' la mia lista:");
     EntryLista en ;
     for (int i=0; i<ListaNodo.size(); i++) {
        en=(EntryLista) ListaNodo.elementAt(i);
        System.out.println (en.IDTabella + " " + en.Locazione + " " +
                            en.Porta + " " + en.Versione + " " + en.Data);
     }
     if ( ListaRifErrati.size() != 0 ) {
       System.out.println ("CA--> Questa e' la mia lista dei riferimenti errati:");
       for (int i=0; i<ListaRifErrati.size(); i++) {
         en = (EntryLista) ListaRifErrati.elementAt(i);
         System.out.println (en.IDTabella + " " + en.Locazione + " " +
                             en.Porta + " " + en.Versione + " " + en.Data);
       }
     } else {
         System.out.println ("CA--> La mia lista dei riferimenti errati e' vuota");
       }
     if (Time2Live > 0) {
       System.out.println ("CA--> Ciao: ora tento la migrazione su un altro nodo... :-)" ) ;
     } else { 
         System.out.println ("CA--> Ciao: e' scaduto il mio tempo... :-|" ) ;
       } 
     MigraSuNodo();
  }
}
