package Agent;

import java.io.*;
import java.net.*;
import java.util.*;
import java.security.*;
import iaik.security.provider.IAIK;
import javax.crypto.*;


// AgentLoader di default
public class DefaultAgentLoader extends Thread implements AgentLoader {

  AgentServer server;
  InputStream iStream;
  OutputStream oStream;
  Socket socket;  

  public DefaultAgentLoader() { super(); }
  
  public void setServer (AgentServer s) { server = s; } 

  public void setInputStream (InputStream i) { iStream = i; }

  public void setOutputStream (OutputStream o) { oStream = o; }

  public void setSocket (Socket sok) { socket = sok; }

  public void run() {
    try {
      // Aggiungo il nuovo security provider
      Provider iaik = new iaik.security.provider.IAIK();
      Security.addProvider(iaik);
      // Istanzio un nuovo cipher
      Cipher c = Cipher.getInstance("DES/ECB8/PKCS5Padding");
      // Inizializzo il cipher per la decifrazione
      c.init (Cipher.DECRYPT_MODE, server.PrivKey );
      // Apro gli ObjectOutputStream nei due sensi
      ObjectInputStream objIStream = new ObjectInputStream (iStream);
      ObjectOutputStream objOStream = new ObjectOutputStream (oStream);
      AgentClassLoader cl = (AgentClassLoader)(getClass().getClassLoader());
      // Recupera il SealObject con l'AgentPacket firmato e cifrato
      SealedObject CryptPack;
      CryptPack = (SealedObject)(objIStream.readObject());
      objOStream.writeObject ( new Boolean(true) );
      System.out.println ("DAL--> SealedObject ricevuto");
      // Recupera l'AgentPack decifrando l'oggetto
      AgentPacket pack = (AgentPacket) CryptPack.getObject(c);
      System.out.println ("DAL--> Ricevuto agente.");
      // Controllo della firma dell'agente
      System.out.println ("DAL--> Controllo la certificazione dell'agente...");
      Signature signature = Signature.getInstance("DSA");
      signature.initVerify (AgentServer.ChiaveCertif);
      String Certif = socket.getInetAddress().getHostAddress() + " 1999";
      System.out.println ("DAL--> Certificazione : " + Certif);
      // Inserimento del certificato da verificare
      for (int i=0; i<Certif.length(); i++) {
        signature.update ((byte) Certif.charAt(i)); 
      }
      if (signature.verify(pack.FirmaBytes)) {
        System.out.println ("DAL--> La firma corrisponde: l'agente viene accettato"); 
        // Aggiorna il database del class loader
        cl.addClassBytes (pack.className, pack.ClassBytes);
        startAgent(pack);
      } else {
          System.out.println ("DAL--> La firma non corrisponde: l'agente viene scartato!!"); 
      }
    } catch ( ClassNotFoundException e ) {
        e.printStackTrace();
    } catch ( IOException ioe ) {
        ioe.printStackTrace();
    } catch ( NoSuchAlgorithmException nsae ) {
        nsae.printStackTrace();
    } catch ( InvalidKeyException ike ) {
        ike.printStackTrace();
    } catch ( SignatureException se ) {
        se.printStackTrace();
    } catch ( NoSuchPaddingException e ) {
        System.out.println ("DAL--> Errore di inizializzazione cipher"); 
    } catch ( BadPaddingException e ) {
        System.out.println ("DAL--> Bad Padding"); 
    } catch ( IllegalBlockSizeException e ) {
        System.out.println ("DAL--> Illegal block size"); 
    }
  }
  
  protected void startAgent (AgentPacket pack) 
                 throws ClassNotFoundException, IOException {
    ByteArrayInputStream byteIStream = new ByteArrayInputStream (pack.AgentBytes);
    ObjectInputStream objIStream = new ObjectInputStream (byteIStream);
    // Questo provocher il caricamento della classe dell'agente
    try {
      Agent agent = (Agent)objIStream.readObject();
      (new AgentExecutor (agent,server)).start();
    } catch ( ClassNotFoundException e ) {
        e.printStackTrace();
    }
   }
}



class AgentExecutor extends Thread {
  Agent agent;
  AgentServer server;
  public AgentExecutor (Agent agent,  AgentServer server) {
    this.agent = agent;
    this.server = server;
  }

  public void run() {
    System.out.println ("AE--> Esecuzione agente : " + agent.AgentName());
    try {
       // Esecuzione dell'agente
       agent.onArrival(server);
    } catch ( AgentException ae ) {
        System.err.println ("AE--> Eccezione AgentException non gestita");
        ae.printStackTrace();
    }
  }
}
