package Agent;

import java.io.*;
import java.util.*;


// Estende la classe vector e aggiunge i metodi di ricerca.
// La classe implementa la interfaccia Serializable poich la Lista pu
// anche essere spedita.
public class ListaTabelle extends java.util.Vector 
                        implements java.io.Serializable {
   
  public ListaTabelle() {						 
    super();
    // Imposto la dimensione massima:
    this.ensureCapacity(25);
  }


  // Metodo di ricerca di una locazione della tabella
  // Restituisce solo la prima occorrenza ...
  public synchronized EntryLista GetLocazione (String Tabella) 
                                      throws Exception {
    EntryLista Risultato;
    for (int Cont=0 ; Cont < this.size(); Cont++) {
      Risultato = (EntryLista) this.elementAt(Cont);
      if ( Risultato.Locazione.compareTo(Tabella) == 0 ) {
	return Risultato;
      }
    }
    throw new Exception("Valore non trovato!");
  }
	

  // Metodo di verifica della presenza di una locazione nella lista
  public synchronized boolean CheckLocazione (String Locazione, int Porta) 
                                                          throws Exception {
    return CheckLocazione (Locazione, Porta, this.size());
  }
 
  public synchronized boolean CheckLocazione (String Locazione, int Porta, int Pos) 
                                                          throws Exception {
    EntryLista Risultato; 
    try {
      for (int Cont = 0; Cont < Pos; Cont++) {
	Risultato = (EntryLista) this.elementAt(Cont);
	if ( ( Risultato.Locazione.compareTo(Locazione) == 0 ) &&
             ( Risultato.Porta == Porta ) ) {
   	  return true;
	}
      }
      return false;
    } catch (Exception e) { 
        throw new Exception ("Errore in CheckLocazione");
    }
  }

  // Metodo che setta il tempo e la versione di una entry relativa ad una tabella
  // Nel caso che la tabella sia mancante aggiunge la entry
  public synchronized int SetTabella (EntryLista entry) {
    // Di default da' l'ordine di aggiornare
    return SetTabella (entry,1);
  }

  public synchronized int SetTabella (EntryLista entry, int Aggiorna) {
    EntryLista Risultato;
    for (int Cont = 0; Cont < this.size(); Cont++) {
      Risultato = (EntryLista) this.elementAt(Cont);
      // Se la locazione della entry corrisponde 
      // e se la versione e' successiva viene aggiornata
      if ( Risultato.Locazione.compareTo(entry.Locazione) == 0 &&	                                   
           Risultato.Porta == entry.Porta &&  
	   Risultato.IDTabella.compareTo(entry.IDTabella) == 0)	{
        // Stessa locazione, porta, tabella 
        if (Risultato.Versione < entry.Versione) {
	  // Se la versione e' piu' recente, aggiorna Data e versione...
	  if ( Aggiorna > 0 ) {
	    ((EntryLista) this.elementAt(Cont)).Versione = entry.Versione;
    	    ((EntryLista) this.elementAt(Cont)).Data = new Date();
	  }
	  return 1;   // fine routine elemento trovato e aggiornato
        } else {      // Se la vesione e' meno aggiornata di quella trovata
	    return -1;
        }
      }
    }
    // Nel caso che la tabella non sia stata trovata
    return 0;
  }

	
  // Metodo di ricerca ed eliminazione di entry scadute
  public synchronized void CheckEntryTime (Date DataScadenza) {
    ListaTabelle dummy = new ListaTabelle();
    CheckEntryTime(DataScadenza,dummy);
  }

  public synchronized void CheckEntryTime (Date DataScadenza,
					   ListaTabelle ListaEliminati) {
    EntryLista Risultato;
    int Cont = 0; 
    while ( Cont < this.size() )	{
       Risultato = (EntryLista) this.elementAt(Cont);
       // Se trova una entry scaduta : viene eliminata dalla ListaTabelle e 
       // inserita nella ListaEliminati
       if ( Risultato.Data.getTime() < DataScadenza.getTime() )	{
	 if (ListaEliminati.SetTabella((EntryLista) this.elementAt(Cont)) == 0)
           ListaEliminati.addElement(this.elementAt(Cont));
  	 this.removeElementAt(Cont);
       } else {
	   Cont++;
       }


    }
  }


  // Metodo di confronto e aggiornamento di liste 
  // Confronto diretto tra due liste
  public synchronized void ConfrontaListe (ListaTabelle Lista) {
    ListaTabelle dummy = new ListaTabelle();
    ConfrontaListe (Lista,dummy);
  }
	
  // Confronto tra due liste, tenendo conto di eventuali entry non valide
  public synchronized void ConfrontaListe (ListaTabelle ListaDaControllare, 
                                           ListaTabelle ListaEliminati) {
    if (ListaDaControllare.equals(null)) 
      return;
    for (int i=0; i<ListaDaControllare.size(); i++) {
      // Ciclo: per ogni entry cerco nella lista..
      switch (this.SetTabella ( (EntryLista)
                               ListaDaControllare.elementAt(i)) )  { 
       case 0 :      // Tabella non trovata
          int Presente = 0;
          if ( (ListaEliminati.size() == 0) || 
               ((Presente = ListaEliminati.SetTabella (
                   (EntryLista) ListaDaControllare.elementAt(i),0)) >= 0) ) {
             // Non e' stata passata la lista dei nodi eliminati -> aggiorno in tutti i casi
             // Una volta recuperata la entry, aggiorno la data e la inserisco nella lista
             // ... Oppure ...
             // Aggiorno solo nel caso che la tabella non sia presente (0) 
             // o ci sia ma piu' vecchia (1)
             EntryLista en = (EntryLista) ListaDaControllare.elementAt(i);
             if (Presente == 1)
               ListaEliminati.DelRifErrati(en);
             EntryLista NuovaEntry = new EntryLista (en.IDTabella, en.Locazione, 
                                                    en.Porta, en.Versione, new Date()); 
	     // Una volta recuperata la entry, aggiorno la data e la inserisco nella lista
             this.CambiaEntry (NuovaEntry);
	  }
          break;						
       case 1 :   // Tabella trovata e aggiornata: non faccio niente
	  break;
       case -1 :  // Tabella trovata ma piu' recente: non faccio niente
	  break; 
      }
    }
  }

 // Confronto INCROCIATO tra due liste, tenendo conto di entry non valide
 public synchronized void ConfrontaListe (ListaTabelle ListaDaControllare, 
                                           ListaTabelle ListaErrati,
                                             ListaTabelle ListaScaduti) {
    // Controllo riferimenti errati ...
    this.DelRifErrati (ListaErrati);
    // Solo una volta tolti gli eventuali riferimenti sbagliati confronta le liste
    this.ConfrontaListe (ListaDaControllare, ListaScaduti);
    // Aggiornamento anche dell'altra lista
    ListaDaControllare.ConfrontaListe (this);
}


  // Metodo per sostituire una generica entry con un riferimento contenente una tabella
  // in modo tale da ottimizzare la lista
  private synchronized void CambiaEntry (EntryLista Nuovaentry) {
    for (int i=0; i<this.size(); i++) {
      EntryLista entry = (EntryLista) this.elementAt(i);
      // Controllo se e' un riferimento allo stesso nodo della nuova entry
      if ( (entry.Locazione.compareTo(Nuovaentry.Locazione) == 0) &&
           (entry.Porta == Nuovaentry.Porta) ) {
        // Se la nuova entry e' soltanto un semplice riferimento senza tabella
        // non occorre inserirlo nella lista
        if (Nuovaentry.IDTabella.compareTo("*") == 0)
          return;
        // Se il riferimento trovato e' un semplice riferimento senza tabella
        // si puo' sostituire la entry
        if (entry.IDTabella.compareTo("*") == 0) {
          this.setElementAt(Nuovaentry,i);
          // System.out.println ("Add:" + Nuovaentry.IDTabella + " " + Nuovaentry.Porta);
          return;
        }
      }
    }
    // Non ci sono riferimenti allo stesso nodo, inserisco la nuova entry
    this.addElement(Nuovaentry);
    // System.out.println ("Add:" + Nuovaentry.IDTabella + " " + Nuovaentry.Porta);
    return;
  }


  
  // Metodo che riceve in ingresso una lista di riferimenti errati di tabelle 
  // Questi riferimenti verranno eliminati dalla lista o trasformati
  // in riferimenti generici.
  public synchronized void DelRifErrati (EntryLista entry) {
  ListaTabelle ListaTemp = new ListaTabelle();
  ListaTemp.addElement(entry);
  this.DelRifErrati(ListaTemp);
  }
  public synchronized void DelRifErrati (ListaTabelle ListaRifErrati) {
     EntryLista Risultato; 
     if (ListaRifErrati.equals(null))  
       return;
     int i = 0;
     while( i<this.size() ) {
        // Ciclo: per ogni entry della lista...
        Risultato = (EntryLista) this.elementAt(i);
        if (ListaRifErrati.SetTabella ((EntryLista) this.elementAt(i),0) <0 ) {
           EntryLista temp = Risultato;
           temp.IDTabella = "*";
           // ...se e' presente un altro riferimento generico a quel nodo
           // lo si cambia, altrimenti si elimina la entry dalla lista
           if ( this.SetTabella(temp,0) == 0 ) {
             ((EntryLista) this.elementAt(i)).IDTabella = "*";
             i++;
           } else {
              this.removeElementAt(i);
           }
        } else {
            i++;
          }
     }
  }

  // Metodo che elimina i riferimenti ad un certo nodo ...
  public synchronized void EliminaRif (String host,int port) {
    EntryLista Risultato; 
    int i = 0;
     if ( (host.compareTo("")==0) || (port<1) ) 
        return;
     while ( i<this.size() ) {
        // Ciclo: per ogni entry della lista...
        Risultato = (EntryLista) this.elementAt(i);
        if ( (Risultato.Locazione.compareTo(host)==0) &&
             (Risultato.Porta == port)  ) {
          // Se e' un riferimento al nodo da eliminare, lo elimino ...
          this.removeElementAt(i);
        } else {
            //... se no, continuo la ricerca
            i++;
          }
     }

  }

  // Metodo che riceve in ingresso un file: analizza ogni linea e compila la lista 
  public synchronized void CompilaLista (File myFile) { 
     String IDTabella;
     String Locazione;
     int Porta;
     int Versione;
      
     try {
       if (myFile.exists() && myFile.canRead()) {
         // Se il file passato esiste ed e' leggibile comincia a lavorare  
         RandomAccessFile file = new RandomAccessFile(myFile, "r");
         while (file.getFilePointer() < file.length()) {
            String entry = file.readLine();
            System.out.println ("Riga letta: " + entry);
            StringTokenizer STentry = new StringTokenizer(entry);
            while (STentry.hasMoreTokens()) {
              IDTabella = STentry.nextToken();
              Locazione = STentry.nextToken();
              Porta = Integer.parseInt(STentry.nextToken());
	      Versione = Integer.parseInt(STentry.nextToken());
      	      System.out.println ("Inserisco elemento: " + IDTabella + " " + Locazione + " " + Porta + " " + Versione);
              EntryLista NuovaEntry = new EntryLista(IDTabella, Locazione, Porta, Versione, new Date());
              this.addElement(NuovaEntry);
            } 
         }
       }
     } catch (Exception e) {
         System.out.println("Errore..." + e.toString());
     }
  }


  // Metodo che riceve un nome di file e restituisce un file contenente i dati della lista 
  public synchronized File ListainFile (String NamemyFile) 
                                             throws Exception {
    File myFile = new File(NamemyFile);
    try {
      RandomAccessFile file = new RandomAccessFile (myFile, "rw");
      for (int i=0; i<this.size(); i++) {
        EntryLista en = (EntryLista) this.elementAt(i);
        file.writeBytes (en.IDTabella + " " + en.Locazione + " " + 
                         en.Porta + " " + en.Versione + " " + en.Data + "\n");
      }
    } catch (Exception e) {
        System.out.println ("Errore..." + e.toString());
    }
    return myFile;
  }

}
