// Classe che genera una coppia di chiavi (pubblica e privata)
// e crea dei file contenenti firme di nodi

// package java.security

import java.io.*;
import java.util.*;
import java.security.*;
import iaik.security.provider.IAIK;
import javax.crypto.*;
import iaik.utils.CryptoUtils;


public class SigFile {

  private static void GeneraChiaveNodo (KeyGenerator GeneratoreChiave, String NomeNodo) {
    Key key;
    try {
      key = GeneratoreChiave.generateKey();
      File filepub;
      File filepriv;
      filepub = new File("PubKey/" + NomeNodo);
      filepriv = new File("PrivKey/" + NomeNodo);
      FileOutputStream fispub = new FileOutputStream(filepub);
      ObjectOutputStream oospub = new ObjectOutputStream(fispub);
      oospub.writeObject(key);
      fispub.close();
      FileOutputStream fispriv = new FileOutputStream(filepriv);
      ObjectOutputStream oospriv = new ObjectOutputStream(fispriv);
      oospriv.writeObject(key);
      fispriv.close();
    } catch (Exception e) {
        System.err.println ("Errore nella generazione delle chiavi! " + e);
      }
  }

  private static byte[] GeneraSignature (Signature signature, PrivateKey privK, String NomeNodo) {
    byte[] realSignature;
    try {
      signature.initSign(privK);
      // Fornisce i dati da firmare per byte
      String Certif = NomeNodo + " 1999";
      for (int i=0; i<Certif.length(); i++) {
        signature.update ((byte) Certif.charAt(i)); 
      }
      // Genera la firma
      realSignature = signature.sign();
      System.out.println ("realSignature " + NomeNodo + "  : " + realSignature);
      File file2 = new File("Certif/" + NomeNodo);
      FileOutputStream fis2 = new FileOutputStream(file2);
      fis2.write(realSignature);
      fis2.close();
      return realSignature;
    } catch (Exception e) {
        System.err.println ("Errore..." + e);
        realSignature = new byte[1];
        return realSignature;
      }
  } 

  private static void VerificaSignature (Signature signature, String NomeNodo, byte[] realSignature) {
 
    String Certif = NomeNodo + " 1999";
    PublicKey pubK;
    try {
      File file = new File("PubKey/pubKey");
      FileInputStream fis = new FileInputStream(file);
      ObjectInputStream ois = new ObjectInputStream(fis);
      pubK = (PublicKey) ois.readObject();
      fis.close();
    } catch (Exception e) {
        System.err.println ("Errore..." + e);
        return;
      }
       
    // Per verificare la firma, l'oggetto Signature viene inizializzato 
    // con la chiave pubblica dell'autorita' di certificazione
    try {
      signature.initVerify(pubK);
      for (int i=0; i<Certif.length(); i++) {
        signature.update ((byte) Certif.charAt(i)); 
      }
      // Verifica della firma
      if (signature.verify(realSignature)) {
        System.out.println("La firma corrisponde"); 
      } else {
          System.out.println("La firma non corrisponde"); 
        }
    } catch (Exception e) {
        System.err.println ("Errore..." + e);
      }
  }


  public static void main (String argv[]) {
    byte[] realSignature;
    ObjectOutputStream oos;
    ObjectInputStream ois;
    String DefaultUrl = "127.0.0.1";
    String user = "certificatore";
    
    try {
      // Tento di creare le directory
      File dir = new File("Certif");
      if ( dir.mkdir() )
        System.out.println ("Creata dir Certif");  
      dir = new File("PrivKey");
      if ( dir.mkdir() )
        System.out.println ("Creata dir PrivKey");
      dir = new File("PubKey");
      if ( dir.mkdir() )
        System.out.println ("Creata dir PubKey");
     } catch ( Exception e ) {
          System.out.println ("Errore nel creare le directory " + e);
        }

    // Istanzio le chiavi
    Key key;
    PublicKey pub;
    PrivateKey priv;
    KeyPair pair;
    // Istanzio il generatore di chiavi
    KeyPairGenerator keygen;
    KeyGenerator kg;
    // Crea un digest
    MessageDigest md;
    // Crea un oggetto Signature per firmare i dati 
    Signature signature;
    
    try {
      Provider iaik = new iaik.security.provider.IAIK();
      Security.addProvider(iaik);

      // Istanzio il digest
      md = MessageDigest.getInstance("SHA");
      String string = DefaultUrl  + user + (new Date()).toString();
      // Aggiorna l'oggetto digest
      md.update(string.getBytes());       
      // Istanzio i generatori di chiavi
      kg = KeyGenerator.getInstance("DES");
      keygen = KeyPairGenerator.getInstance("DSA");
      keygen.initialize(512,new SecureRandom(md.digest()));
      // Genero la coppia di chiavi del certificatore
      pair = keygen.generateKeyPair();
      System.out.println ("Public-key:  " + pair.getPublic().getEncoded() + "   Private-key: " + 
                          pair.getPrivate().getEncoded() + "\n");
      System.out.println ("Public-key:  " + pair.getPublic().toString());
      System.out.println ("Private-key: " + pair.getPrivate().toString());   
      // Estrae chiave privata 
      priv = pair.getPrivate();
      // Estrae chiave pubblica
      pub = pair.getPublic();
      // Istanzio l'oggetto signature
      signature = Signature.getInstance("DSA");
    } catch ( NoSuchAlgorithmException nsae ) {
        System.out.println ("Algoritmo inesistente!! " + nsae);
        return;
    } catch ( Exception e ) {
        System.out.println ("Errore !! " + e);
        return;
      }

    // Scrivo la chiave pubblica nel file "PubKey/pubKey"
    try {
      File file3 = new File("PubKey/pubKey");
      FileOutputStream fis3 = new FileOutputStream(file3);
      oos = new ObjectOutputStream(fis3);
      oos.writeObject(pub);
      fis3.close();
    } catch ( Exception e ) {
        System.out.println ("Errore nel salvare la chiave pubblica " + e);
        return;
      }
   
    String Comando = "";
    while ( (Comando.compareTo("exit") != 0) ) {
      System.out.println ("Inserire l'indirizzo IP del nodo da certificare");
      System.out.println ("Inserire exit per uscire");
      BufferedReader dis = new BufferedReader (new InputStreamReader(System.in));
      try {
        Comando = dis.readLine();
      } catch ( IOException e ) {
          System.out.println("Errore : " + e.toString() );
        }
      if ( (Comando.length() > 0) &&
           (Comando.compareTo("exit") != 0) ) {
        GeneraChiaveNodo (kg, Comando);
        realSignature = GeneraSignature (signature, priv, Comando);
        VerificaSignature (signature, Comando, realSignature);
      }
    System.out.println();
    }
    System.out.println ("Fine esecuzione... Ciao :-)");
  }
    
}


