// Agente di ricerca

import Agent.*; 
import java.io.*;
import java.util.*;
import java.net.*;


public class SearchAgent extends AgentEx {
  
  protected static String host;
  protected static int port;
  // Lista dei "riferimenti conosciuti"
  protected ListaTabelle ListaNodo;
  // Lista dei riferimenti errati alle tabelle 
  protected ListaTabelle ListaRifErrati;
  // Array contenente le tabelle da ricercare
  protected Hashtable ArrayTabelle;
  
  // Variabili per il Time to Live dell'agente
  protected int Time2Live;
  // Indice della ricerca
  protected int NNodiVisitati;
  // Informazioni sul Nodo che ospita l'agente
  protected String HostLoc;
  protected int HostPort;
  // Variabile di stato dell'agente
  protected int Stato;	
  // Variabile di LOG: informazioni sugli spostamenti dell'agente
  protected String Log = "";
  // Informazioni su chi ha mandato l'agente 
  protected String HostPrincipal;
  protected int PortPrincipal;
	

  public SearchAgent (String s, int p, Hashtable Arr, int NuovoStato, int NuovoNNodiVisitati, 
                      byte Firma[], Hashtable kl) {
    super ( "Search Agent" );   // Agente di ricerca
    // Nodo di origine dell'agente: quello da cui e'
    // stato fatto partire dal QueryAgent..
    HostPrincipal = s;
    PortPrincipal = p;
    // Istanzio liste, ArrayTabelle e altre variabili di stato 
    ListaNodo = new ListaTabelle();
    ListaRifErrati = new ListaTabelle();
    KeyList = new Hashtable();
    KeyList = kl;
    ArrayTabelle = Arr; 
    NNodiVisitati = NuovoNNodiVisitati;
    Stato = NuovoStato;
    Time2Live = 20;
    HostLoc = "";
    HostPort = 0;
    // Settaggio della la firma....
    FirmaBytes = Firma;
  }


  // Metodo per aggiornare le liste	
  public void SetListeNodo (ListaTabelle SerLista, ListaTabelle SerLe, Hashtable SerKl) {

    // Se la dimensione di ListaNodo e' = 0, significa che  
    // il SearchAgent e' stato appena lanciato dal QueryAgent, 
    // quindi bisogna fare l'assegnamento della lista dei riferimenti errati
    if ( ListaNodo.size() == 0 )
      ListaRifErrati = SerLe;
    System.out.println ("Agent--> Liste ricevute: controllo e aggiornamento...");
    // Controllo ed elimino riferimenti errati al nodo del server..
    for (int i=0; i<ListaNodo.size(); i++) {
    EntryLista en = (EntryLista) ListaNodo.elementAt(i);
       if ( (en.IDTabella.compareTo("*") != 0) &&
            (en.Locazione.compareTo(HostLoc) == 0 ) &&
            (en.Porta == HostPort ) ) {
         // Nel caso che il riferimento sia AD UNA TABELLA su questo server 
         // ne controllo la validita' ed eventualmente lo elimino...
         if ( SerLista.SetTabella(en,0) == 0 ) {
           // Aggiunta del riferimento errato nella relativa lista...
           if ( ListaRifErrati.SetTabella((EntryLista) ListaNodo.elementAt(i)) == 0 )
             ListaRifErrati.addElement ((EntryLista) ListaNodo.elementAt(i));
           System.out.println ("Agent--> Eliminata la entry errata: ");
           System.out.println (en.IDTabella + " " + en.Locazione + " " + 
                               en.Porta + " " + en.Versione + " " + en.Data); 
           EntryLista temp = new EntryLista ("*", en.Locazione, en.Porta, en.Versione, en.Data);
           // ...se e' presente un altro riferimento generico a quel nodo
           // lo si cambia, altrimenti si elimina la entry dalla lista
           if (ListaNodo.SetTabella(temp) == 0) {
             ListaNodo.setElementAt(temp,i);
           } else {
               ListaNodo.removeElementAt(i);
               // Aggiornamento eventuale del puntatore di migrazione...
               if ( i <= NNodiVisitati )
                 NNodiVisitati--;
             }
           i--;  // Avendo eliminato una entry decremento il contatore
         }
       }
    }

    // Controllo incrociato delle liste
    SerLista.ConfrontaListe (ListaNodo, ListaRifErrati, SerLe);

    // Aggiornamento incrociato delle liste dei riferimenti errati
    SerLe.ConfrontaListe (ListaRifErrati);
    // Elimino riferimenti errati inutili al server stesso
    SerLe.EliminaRif (HostLoc, HostPort);

    // Aggiornamento delle chiavi
    for (int i=0; i<SerLista.size(); i++) {
      try {
        EntryLista ent = (EntryLista) SerLista.elementAt(i);
        SerKl.put (ent.Locazione, KeyList.get(ent.Locazione));
      } catch (Exception e) { }
    }
       KeyList = SerKl;
  }


  // Metodo di inizio attivit dell'agente
  protected void execute() throws AgentException {
    System.out.println ("Salve sono l'agente " + AgentName());
    System.out.println ("SA--> Inizio la ricerca delle tabelle non trovate :-|");
    MigraSuNodo();	
  }


  // Metodo di migrazione dell'agente
  protected void MigraSuNodo() {
    // Se non ci sono piu' nodi da visitare 
    // o il Time to Live  scaduto termina  
    while ((ListaNodo.size() > NNodiVisitati) && (Time2Live > 0)) { 
      // Tentativo di migrazione
      System.out.println ("SA--> Tento la migrazione..."); 
      EntryLista en;
      en = (EntryLista) ListaNodo.elementAt(NNodiVisitati);
      try {
        if (ListaNodo.CheckLocazione(en.Locazione, en.Porta, NNodiVisitati)) {
          // Non migra su un nodo gia' visitato, ma
          // incrementa comunque il numero di nodi visitati..
          NNodiVisitati++;
          System.out.println ("SA--> Su questo nodo ci sono gia' stato... : " + en.Locazione +
                              " : " + en.Porta);
        } else {
            String OldLoc = HostLoc;
            int OldPort = HostPort;
            try {
              Log = Log  + "SA-> " + " " + Stato + " " + en.Locazione + " : " + en.Porta + "\n";
              System.out.println ("Log: <stato>  <locazione>  <porta>");
              System.out.println (Log);
              // Salvataggio delle informazioni sul nodo dove si sta migrando
              HostLoc = en.Locazione;
              HostPort = en.Porta;
              // Tentativo di migrazione
              migrate(en.Locazione, en.Porta);
              return;         
            } catch (AgentException ae) {
                // Reinserico i valori corretti
                HostPort = OldPort ;
                HostLoc = OldLoc ;
                Log = Log + "SA-> Tentativo fallito..." + "\n";
                // Eliminazione della entry del nodo che ha causato l'errore
                ListaNodo.removeElementAt(NNodiVisitati);
                System.out.print ("SA--> Eliminata la entry:"); 
                System.out.println (en.IDTabella + " " + en.Locazione + " " + 
                                    en.Porta + " " + en.Versione + " " + en.Data);
            }
          }
      } catch (Exception e) { 
          System.out.println ("SA--> Errore durante il check della lista");   
      }
    }
    System.out.println ("SA--> Siamo arrivati in fondo alla lista... ");
    System.out.println ("SA--> Bisogna comunicare al QueryAgent che la ricerca e' fallita..");
    // Fallimento ricerca: ritorno dal QueryAgent... ci prova per 3 volte..
    int Tentativo = 1;
    while ( Tentativo < 4 ) {
        try {
          // Settaggio dello stato di fine ricerca con insuccesso
   	  Stato = 4;  
          System.out.println ("SA--> Tentativo di ritorno a " + HostPrincipal + ":" + PortPrincipal );     
          HostLoc = HostPrincipal;
          HostPort = PortPrincipal;
          migrate(HostPrincipal,PortPrincipal);
          Tentativo = 4;
        } catch (AgentException ae) {
            // Migrazione di ritorno fallita
            Tentativo++;
            if ( Tentativo <= 3 ) {
             System.out.println ("SA--> Migrazione di ritorno numero " + Tentativo + " fallita :-(( ");
             try {
               this.sleep(10000*Tentativo);
             } catch ( InterruptedException  ie ) {
                System.out.println ("SA--> Attesa interrotta "); 
               }
	    } else {  
                System.out.println ("SA--> Migrazione di ritorno fallita definitivamente !!!:-(( ");
              }
        } 
    } 
  }   


  // Metodo chiamato all'arrivo su un sito remoto
  public void onArrival(AgentServer server) throws AgentException {

     // Aggiornamento delle liste a carico dell'agente...
     System.out.println ("SA--> Liste date all'agente per l'aggiornamento ");
     SetListeNodo (server.Lista, server.ListaEliminati, server.KeyList);
     // Si da' la firma all'agente...
     SetFirma (server.FirmaNodo);

     // Incremento del numero di nodi visitati
     NNodiVisitati++;
     // Decremento del Time to live
     Time2Live--;
     System.out.println ("Salve, sono l'agente " + AgentName() + " appena arrivato..." ) ;
     // Visualizzazione delle liste sul nodo di arrivo
     System.out.println ("SA--> Questa e' la mia lista(" + NNodiVisitati +"):");
     EntryLista en;
     for (int i=0; i<ListaNodo.size(); i++) {
        en = (EntryLista) ListaNodo.elementAt(i);
	System.out.println (en.IDTabella + " " + en.Locazione + " " +
                            en.Porta + " " + en.Versione + " " + en.Data);
     }
     if ( ListaRifErrati.size() != 0 ) {
       System.out.println ("SA--> Questa e' la mia lista dei riferimenti errati: ");
       for (int i=0; i<ListaRifErrati.size(); i++) {
          en = (EntryLista) ListaRifErrati.elementAt(i);
	  System.out.println (en.IDTabella + " " + en.Locazione + " " +
                              en.Porta + " " + en.Versione + " " + en.Data);
       }
     } else {
         System.out.println ("SA--> La mia lista dei riferimenti errati e' vuota");
       }
     try {
       LeggiStato();
     } catch (Exception e) {
         System.out.println ("SA--> Errore di stato: " + Stato + "  " + e.toString());
     }
  }


  // Metodo di analisi dello stato di esecuzione dell'agente
  protected void LeggiStato() throws Exception {
     switch (Stato) {
        case 1:
          RicercaTabelle();
          break;
        case 2:
          VerificaTabelle();
          break;
       	case 3: 
        case 4:
          FineRicerca();
	  break;
        default:
          throw new Exception ("SA--> Stato non corretto!!"); 
     }
  }
  

  // Metodo chiamato alla fine della ricerca delle tabelle 
  protected void FineRicerca() {
    Log = Log  + "SA-> " + " " + Stato + " " + HostLoc + " : " + HostPort + "\n";
    // Comunicazione con QueryAgent
    try {
       Socket SocketRitorno = new Socket (HostPrincipal, 60000);
       ObjectOutputStream OUT = new ObjectOutputStream (SocketRitorno.getOutputStream());
       if ( Stato == 4) {
         OUT.writeObject (new Boolean(false));
       } else {
             OUT.writeObject (new Boolean(true));
           }
       OUT.writeObject (ListaNodo);
       OUT.writeObject (ListaRifErrati);
       OUT.writeObject (KeyList);
       OUT.writeObject (Log);
       OUT.close();
       SocketRitorno.close(); 
    } catch (UnknownHostException uhe) {
        System.out.println ("SA--> Non trovo l'host!! " + uhe.toString());
        System.out.println ("SA--> Termino l'esecuzione :-(( ");
    } catch (IOException ioe) {
        System.out.println ("SA--> Errore di rete!!! " + ioe.toString());
        System.out.println ("SA--> Termino l'esecuzione :-(( ");
    }    
    Time2Live = 0;
  }


  // Metodo di verifica della presenza di tutte le tabelle interessate dalla query
  protected void RicercaTabelle() {
     int i = 0;  
     Hashtable ArrayTemp = new Hashtable();
     while ( i < ListaNodo.size() ) { 
	EntryLista en;
        en = (EntryLista) ListaNodo.elementAt(i);
        if (ArrayTabelle.containsKey(en.IDTabella)) { 
          ArrayTemp.put (en.IDTabella, en.IDTabella);
        }
     i++;
     }
     if (ArrayTemp.size() == ArrayTabelle.size()) {
       System.out.println ("SA--> Verifica effettiva presenza delle tabelle nei nodi...");
       // Settaggio dello stato di inizio verifica
       Stato = 2;  
       // Verifica se sul nodo corrente ci sono delle tabelle che interessano
       CercaTabLocali (HostLoc, HostPort);
       MigraSuNodo2();
     } else {
         System.out.println ("SA--> Non trovate tutte le tabelle, proseguo la ricerca...:-(");
         MigraSuNodo();
       }
  }


  // Metodo di migrazione chiamato durante la verifica della presenza effettiva delle tabelle
  protected void MigraSuNodo2 () {
    // Se il Time to Live  scaduto termina  
    while ( (Time2Live > 0) && (ArrayTabelle.size() > 0) ) { 
      // Tentativo di migrazione
      System.out.println ("SA--> Tento la migrazione..."); 
      EntryLista en;
      // Se la tabella e' presente nell'ArrayTabelle ed e' locale:
      for (int i=0; (i<ListaNodo.size()) ; i++) {
	 en = (EntryLista) ListaNodo.elementAt(i);
         try {
           if ( (ArrayTabelle.containsKey(en.IDTabella)) ) { 
             if ( !ListaNodo.CheckLocazione(en.Locazione, en.Porta, NNodiVisitati) ) {
               // Migra solo su un nodo dove e' presente una tabella che interessa 
               // e che non e' stato ancora visitato
               String OldLoc = HostLoc;
               int OldPort = HostPort;
               try {
                 Log = Log + "SA-> " + " " + Stato + " " + en.Locazione + " : " +  en.Porta + "\n";
                 System.out.println ("Log: <stato>  <locazione>  <porta>");
                 System.out.println (Log);
                 // Salvataggio delle informazioni sul nodo dove si sta migrando
                 HostLoc = en.Locazione;
                 HostPort = en.Porta;
                 // Tentativo di migrazione
                 migrate (en.Locazione, en.Porta);
                 return;         
               } catch (AgentException ae) {
                   // Reinserico i valori corretti
                   HostPort = OldPort ;
                   HostLoc = OldLoc ;
                   Log = Log + "SA-> Tentativo fallito..." + "\n";
                   // Elimina la entry del nodo che ha causato l'errore adeguando, se e' il caso, NNodiVisitati
                   if ( i <= NNodiVisitati ) {
                     NNodiVisitati--;
                   }
                   ListaNodo.removeElementAt(i);
                   i--;
                   System.out.print ("SA--> Eliminata la entry:"); 
                   System.out.println (en.IDTabella + " " + en.Locazione +  " " + en.Porta + " "
                                       + en.Versione + " " + en.Data);
               }
             } else {
                 // Il nodo e' gia' stato controllato, il riferimento alla
                 // tabella e' corretto
                 ArrayTabelle.remove (en.IDTabella);
               }
           }
         } catch (Exception e) { 
             System.out.println ("SA--> Errore durante il check della lista");   
           }
        } 
        if (ArrayTabelle.size() > 0) {
          System.out.println ("SA--> Non trovate tutte le tabelle, proseguo la ricerca...:-(");
          Stato = 1;
          MigraSuNodo();
          Time2Live = 0;
        }
    } 
    if ( ArrayTabelle.size() == 0 ) {
      System.out.println ("SA--> Verificata la presenza di tutte le tabelle, ritorno dal QueryAgent...:-)");       
      int Tentativo = 1; 
      while ( Tentativo < 4 ) {
        try {
          // Settaggio dello stato di fine ricerca con successo
   	  Stato = 3;  
          System.out.println ("SA--> Tentativo di ritorno a " + HostPrincipal + ":" + PortPrincipal );     
          HostLoc = HostPrincipal;
          HostPort = PortPrincipal;
          migrate (HostPrincipal, PortPrincipal);
          Tentativo = 4;
        } catch (AgentException ae) {
           // Migrazione di ritorno fallita
           Tentativo++;
           System.out.println ("SA--> Migrazione di ritorno numero " + Tentativo + " fallita :-(( ");
           try {
             this.sleep(10000*Tentativo);
           } catch ( InterruptedException  ie ) {
              System.out.println ("SA--> Attesa interrotta "); 
             }
           if ( Tentativo == 4 ) {
             System.out.println ("SA--> Migrazione di ritorno fallita definitivamente !!!:-(( ");
           }
        } 
      } 
      Time2Live = 0;
    } 
  }   


  // Metodo di verifica della presenza di tabelle interessate dalla query
  protected void VerificaTabelle() {
     EntryLista en;
     // Selezione entry corrispondente al nodo in cui ci si trova
     if ( CercaTabLocali(HostLoc,HostPort) ) { 
       // Continua la verifica...
       MigraSuNodo2();
     } else {
         // Siamo stati portati qui da un riferimento errato...
         // Inizio di nuovo la ricerca...
         Stato = 1;
         MigraSuNodo();    
       }            
  }


  // Metodo di ricerca di tabelle su un nodo
  protected boolean CercaTabLocali (String Locazione, int Porta) {
    // Per ogni entry della lista tabelle, si controlla se e' presente nell'ArrayTabelle 
    EntryLista en;
    // (Il settaggio di Modifica e' invertito rispetto al QueryAgent)
    boolean Modifica = false;
    System.out.println ("SA--> Ricerca su " + Locazione + ":" + Porta );
    for (int i=0; i<ListaNodo.size(); i++) {
      en = (EntryLista) ListaNodo.elementAt(i);
      // Se la tabella e' presente nell'ArrayTabelle ed e' locale
      // elimina il riferimento alla tabella da ArrayTabelle
      if ( (ArrayTabelle.containsKey(en.IDTabella)) &&
           (en.Porta == Porta)  &&
           (en.Locazione.compareTo(Locazione) == 0) )  {
        ArrayTabelle.remove (en.IDTabella);
        Modifica = true;
      }
    }
    return Modifica;
  }
}
